class Spec < ActiveRecord::Base
  belongs_to :user
  acts_as_ferret
  
  ALL_FIELDS = %w(first_name last_name occupation gender birthdate city state zip_code)
  STRING_FIELDS = %w(first_name last_name occupation city state)
  VALID_GENDERS = ["Mężczyzna", "Kobieta"]
  START_YEAR = 1900
  VALID_DATES = DateTime.new(START_YEAR)..DateTime.now
  ZIP_CODE_LENGTH = 5
  
  HUMANIZED_ATTRIBUTES = {
    :first_name => "Imię",
    :last_name => "Nazwisko",
    :occupation => "Zawód",
    :gender => "Płeć",
    :birthdate => "Data urodzenia",
    :city => "Miasto",
    :state => "Województwo",
    :zip_code => "Kod pocztowy",
    :spec => "specyfikacji"
  }
  
  def self.human_attribute_name(attr)
    HUMANIZED_ATTRIBUTES[attr.to_sym] || super
  end
  
  
  validates_length_of STRING_FIELDS,
                      :maximum => DB_STRING_MAX_LENGTH
  validates_inclusion_of :gender,
                         :in => VALID_GENDERS,
                         :allow_nil => true,
                         :message => "należy wybrać mężczyzna albo kobieta"
  validates_inclusion_of :birthdate,
                         :in => VALID_DATES,
                         :allow_nil => true,
                         :message => "jest nieprawidłowa"
  validates_format_of :zip_code, 
                      :with => /(^$|^[0-9]{#{ZIP_CODE_LENGTH}}$)/,
                      :message => "musi być liczbą pięciocyfrową."

  # Zwraca pełne dane osobowe (imię i nazwisko).
  def full_name
    [first_name, last_name].join(" ")
  end

  # Zwraca rozsądnie sformatowany łańcuch z lokalizacją użytkownika
  def location
    if not zip_code.blank? and (city.blank? or state.blank?)
      lookup = GeoDatum.find_by_zip_code(zip_code)
      if lookup
        self.city = lookup.city.capitalize_each if city.blank?
        self.state = lookup.state if state.blank?
      end
    end
    [city, state, zip_code].join(" ")
  end

  
  # Zwraca wiek obliczany na podstawie daty urodzenia
  def age
    return if birthdate.nil?
    today = Date.today
    if (today.month > birthdate.month) or
      (today.month == birthdate.month and today.day >= birthdate.day)
      # Urodziny już były w tym roku
      today.year - birthdate.year
    else
      today.year - birthdate.year - 1
    end
  end

  # Wyszukiwanie wg wieku, płci i miejsca pobytu
  def self.find_by_asl(params)
    where =[]
    # Ustawia ograniczenia wieku jako limity zakresu dat urodzenia w SQL.
    unless params[:min_age].blank?
      where << "ADDDATE(birthdate, INTERVAL :min_age YEAR) < CURDATE()"
    end
    unless params[:max_age].blank?
    where << "ADDDATE(birthdate, INTERVAL :max_age+1 YEAR) > CURDATE()"
    end
    # Ustawia ograniczenie płci w SQL.
    where << "gender = :gender" unless params[:gender].blank?
    # Ustawia ograniczenia odległości w SQL.
    zip_code = params[:zip_code]
    unless zip_code.blank? and params[:miles].blank?
      location = GeoDatum.find_by_zip_code(zip_code)
      distance = sql_distance_away(location)
      where << "#{distance} <= :miles"
     end  
    if where.empty?
      []
    else
      find(:all,
         :joins => "LEFT JOIN geo_data ON geo_data.zip_code = specs.zip_code",
         :conditions => [where.join(" AND "), params], 
         :order => "last_name, first_name")
    end
  end

  
private  

  # Zwraca SQL dla odległości między lokalizacja w specyfikacji a podanym punktem.
  # Więcej ifnormacji na temat wzoru znajdziesz pod adresem http://en.wikipedia.org/wiki/Haversine_formula
  def self.sql_distance_away(point)
    h = "POWER(SIN((RADIANS(latitude - #{point.latitude}))/2.0),2) + " +
    "COS(RADIANS(#{point.latitude})) * COS(RADIANS(latitude)) * " +
    "POWER(SIN((RADIANS(longitude - #{point.longitude}))/2.0),2)"
    r = 3956 # Promień Ziemi w milach
    "2 * #{r} * ASIN(SQRT(#{h}))"
  end


end
